#pragma once

//TODO: Popis MFRC522, jednotlivych rfid karet a pouziti 

//Knihovna vyuziva SPI komunikaci

//reader - PCD (Proximity Coupling Device)
//card   - PICC (Proximity Integrated Circuit Card)

#include <avr/io.h>
#include <stdint.h>
#include <stddef.h>
#include <string.h>


#include "spi.h"
#include "wait.h"



class MFRC522_class
{
	//Definice
	public:
		
		//MFRC522 registry
		//Vsechny adresy jsou posunuty o 1 bit vlevo (datasheet 8.1.2.3)
		enum reader_register
		{
			// Page 0: Command and status
			//						  0x00			// reserved for future use
			CommandReg				= 0x01 << 1,	// starts and stops command execution
			ComIEnReg				= 0x02 << 1,	// enable and disable interrupt request control bits
			DivIEnReg				= 0x03 << 1,	// enable and disable interrupt request control bits
			ComIrqReg				= 0x04 << 1,	// interrupt request bits
			DivIrqReg				= 0x05 << 1,	// interrupt request bits
			ErrorReg				= 0x06 << 1,	// error bits showing the error status of the last command executed
			Status1Reg				= 0x07 << 1,	// communication status bits
			Status2Reg				= 0x08 << 1,	// receiver and transmitter status bits
			FIFODataReg				= 0x09 << 1,	// input and output of 64 byte FIFO buffer
			FIFOLevelReg			= 0x0A << 1,	// number of bytes stored in the FIFO buffer
			WaterLevelReg			= 0x0B << 1,	// level for FIFO underflow and overflow warning
			ControlReg				= 0x0C << 1,	// miscellaneous control registers
			BitFramingReg			= 0x0D << 1,	// adjustments for bit-oriented frames
			CollReg					= 0x0E << 1,	// bit position of the first bit-collision detected on the RF interface
			//						  0x0F			// reserved for future use
			
			// Page 1:Command
			// 						  0x10			// reserved for future use
			ModeReg					= 0x11 << 1,	// defines general modes for transmitting and receiving
			TxModeReg				= 0x12 << 1,	// defines transmission data rate and framing
			RxModeReg				= 0x13 << 1,	// defines reception data rate and framing
			TxControlReg			= 0x14 << 1,	// controls the logical behavior of the antenna driver pins TX1 and TX2
			TxASKReg				= 0x15 << 1,	// controls the setting of the transmission modulation
			TxSelReg				= 0x16 << 1,	// selects the internal sources for the antenna driver
			RxSelReg				= 0x17 << 1,	// selects internal receiver settings
			RxThresholdReg			= 0x18 << 1,	// selects thresholds for the bit decoder
			DemodReg				= 0x19 << 1,	// defines demodulator settings
			// 						  0x1A			// reserved for future use
			// 						  0x1B			// reserved for future use
			MfTxReg					= 0x1C << 1,	// controls some MIFARE communication transmit parameters
			MfRxReg					= 0x1D << 1,	// controls some MIFARE communication receive parameters
			// 						  0x1E			// reserved for future use
			SerialSpeedReg			= 0x1F << 1,	// selects the speed of the serial UART interface
			
			// Page 2: Configuration
			// 						0x20			// reserved for future use
			CRCResultRegH			= 0x21 << 1,	// shows the MSB and LSB values of the CRC calculation
			CRCResultRegL			= 0x22 << 1,
			// 						  0x23			// reserved for future use
			ModWidthReg				= 0x24 << 1,	// controls the ModWidth setting?
			// 						  0x25			// reserved for future use
			RFCfgReg				= 0x26 << 1,	// configures the receiver gain
			GsNReg					= 0x27 << 1,	// selects the conductance of the antenna driver pins TX1 and TX2 for modulation
			CWGsPReg				= 0x28 << 1,	// defines the conductance of the p-driver output during periods of no modulation
			ModGsPReg				= 0x29 << 1,	// defines the conductance of the p-driver output during periods of modulation
			TModeReg				= 0x2A << 1,	// defines settings for the internal timer
			TPrescalerReg			= 0x2B << 1,	// the lower 8 bits of the TPrescaler value. The 4 high bits are in TModeReg.
			TReloadRegH				= 0x2C << 1,	// defines the 16-bit timer reload value
			TReloadRegL				= 0x2D << 1,
			TCounterValueRegH		= 0x2E << 1,	// shows the 16-bit timer value
			TCounterValueRegL		= 0x2F << 1,
			
			// Page 3:Test Registers
			// 						  0x30			// reserved for future use
			TestSel1Reg				= 0x31 << 1,	// general test signal configuration
			TestSel2Reg				= 0x32 << 1,	// general test signal configuration
			TestPinEnReg			= 0x33 << 1,	// enables pin output driver on pins D1 to D7
			TestPinValueReg			= 0x34 << 1,	// defines the values for D1 to D7 when it is used as an I/O bus
			TestBusReg				= 0x35 << 1,	// shows the status of the internal test bus
			AutoTestReg				= 0x36 << 1,	// controls the digital self test
			VersionReg				= 0x37 << 1,	// shows the software version
			AnalogTestReg			= 0x38 << 1,	// controls the pins AUX1 and AUX2
			TestDAC1Reg				= 0x39 << 1,	// defines the test value for TestDAC1
			TestDAC2Reg				= 0x3A << 1,	// defines the test value for TestDAC2
			TestADCReg				= 0x3B << 1		// shows the value of ADC I and Q channels
			// 						  0x3C			// reserved for production tests
			// 						  0x3D			// reserved for production tests
			// 						  0x3E			// reserved for production tests
			// 						  0x3F			// reserved for production tests
		};
		
		
		//MFRDC522 prikazy
		enum reader_command
		{
			READER_Idle					= 0x00,		// no action, cancels current command execution
			READER_Mem					= 0x01,		// stores 25 bytes into the internal buffer
			READER_GenerateRandomID		= 0x02,		// generates a 10-byte random ID number
			READER_CalcCRC				= 0x03,		// activates the CRC coprocessor or performs a self test
			READER_Transmit				= 0x04,		// transmits data from the FIFO buffer
			READER_NoCmdChange			= 0x07,		// no command change, can be used to modify the CommandReg register bits without affecting the command, for example, the PowerDown bit
			READER_Receive				= 0x08,		// activates the receiver circuits
			READER_Transceive 			= 0x0C,		// transmits data from FIFO buffer to antenna and automatically activates the receiver after transmission
			READER_MFAuthent 			= 0x0E,		// performs the MIFARE standard authentication as a READER
			READER_SoftReset			= 0x0F		// resets the MFRC522
		};
		
		
		//Prikazy pro rfid karty
		enum card_command
		{
			// The commands used by the reader to manage communication with several PICCs (ISO 14443-3, Type A, section 6.4)
			CARD_CMD_REQA			= 0x26,		// REQuest command, Type A. Invites CARDs in state IDLE to go to READY and prepare for anticollision or selection. 7 bit frame.
			CARD_CMD_WUPA			= 0x52,		// Wake-UP command, Type A. Invites CARDs in state IDLE and HALT to go to READY(*) and prepare for anticollision or selection. 7 bit frame.
			CARD_CMD_CT				= 0x88,		// Cascade Tag. Not really a command, but used during anti collision.
			CARD_CMD_SEL_CL1		= 0x93,		// Anti collision/Select, Cascade Level 1
			CARD_CMD_SEL_CL2		= 0x95,		// Anti collision/Select, Cascade Level 1
			CARD_CMD_SEL_CL3		= 0x97,		// Anti collision/Select, Cascade Level 1
			CARD_CMD_HLTA			= 0x50,		// HaLT command, Type A. Instructs an ACTIVE CARD to go to state HALT.
			// The commands used for MIFARE Classic (from http://www.nxp.com/documents/data_sheet/MF1S503x.pdf, Section 9)
			// Use PCD_MFAuthent to authenticate access to a sector, then use these commands to read/write/modify the blocks on the sector.
			// The read/write commands can also be used for MIFARE Ultralight.
			CARD_CMD_MF_AUTH_KEY_A	= 0x60,		// Perform authentication with Key A
			CARD_CMD_MF_AUTH_KEY_B	= 0x61,		// Perform authentication with Key B
			CARD_CMD_MF_READ		= 0x30,		// Reads one 16 byte block from the authenticated sector of the CARD. Also used for MIFARE Ultralight.
			CARD_CMD_MF_WRITE		= 0xA0,		// Writes one 16 byte block to the authenticated sector of the CARD. Called "COMPATIBILITY WRITE" for MIFARE Ultralight.
			CARD_CMD_MF_DECREMENT	= 0xC0,		// Decrements the contents of a block and stores the result in the internal data register.
			CARD_CMD_MF_INCREMENT	= 0xC1,		// Increments the contents of a block and stores the result in the internal data register.
			CARD_CMD_MF_RESTORE		= 0xC2,		// Reads the contents of a block into the internal data register.
			CARD_CMD_MF_TRANSFER	= 0xB0,		// Writes the contents of the internal data register to a block.
			// The commands used for MIFARE Ultralight (from http://www.nxp.com/documents/data_sheet/MF0ICU1.pdf, Section 8.6)
			// The CARD_CMD_MF_READ and CARD_CMD_MF_WRITE can also be used for MIFARE Ultralight.
			CARD_CMD_UL_WRITE		= 0xA2		// Writes one 4 byte page to the PICC.
		};
		
		
		//MIFARE konstanty
		enum mifare_misc
		{
			MF_ACK					= 0xA,		// The MIFARE Classic uses a 4 bit ACK/NAK. Any other value than 0xA is NAK.
			MF_KEY_SIZE				= 6			// A Mifare Crypto1 key is 6 bytes.
		};
		
		
		//Typy karet, ktere umime detekovat (po pridani dalsich je treba updatovat card_get_type_name() )
		enum card_type
		{
			CARD_TYPE_UNKNOWN		= 0,
			CARD_TYPE_ISO_14443_4	= 1,	// CARD compliant with ISO/IEC 14443-4
			CARD_TYPE_ISO_18092		= 2, 	// CARD compliant with ISO/IEC 18092 (NFC)
			CARD_TYPE_MIFARE_MINI	= 3,	// MIFARE Classic protocol, 320 bytes
			CARD_TYPE_MIFARE_1K		= 4,	// MIFARE Classic protocol, 1KB
			CARD_TYPE_MIFARE_4K		= 5,	// MIFARE Classic protocol, 4KB
			CARD_TYPE_MIFARE_UL		= 6,	// MIFARE Ultralight or Ultralight C
			CARD_TYPE_MIFARE_PLUS	= 7,	// MIFARE Plus
			CARD_TYPE_TNP3XXX		= 8,	// Only mentioned in NXP AN 10833 MIFARE Type Identification Procedure
			CARD_TYPE_NOT_COMPLETE	= 255	// SAK indicates UID is not complete.
		};
		
		
		//Navratove hodnoty metod teto tridy (po pridani dalsich se musi updatovat get_status_code_name() )
		enum status_code
		{
			STATUS_OK				= 1,	// Success
			STATUS_ERROR			= 2,	// Error in communication
			STATUS_COLLISION		= 3,	// Collission detected
			STATUS_TIMEOUT			= 4,	// Timeout in communication.
			STATUS_NO_ROOM			= 5,	// A buffer is not big enough.
			STATUS_INTERNAL_ERROR	= 6,	// Internal error in the code. Should not happen ;-)
			STATUS_INVALID			= 7,	// Invalid argument.
			STATUS_CRC_WRONG		= 8,	// The CRC_A does not match
			STATUS_MIFARE_NACK		= 9		// A MIFARE PICC responded with NAK.
		};
		
		
		//Definice gainu
		enum gain_code
		{
			GAIN_18dB	= 0x00,
			GAIN_23dB	= 0x10,
			GAIN_18dB2	= 0x20,
			GAIN_23dB2	= 0x30,
			GAIN_33dB	= 0x40,
			GAIN_38dB	= 0x50,
			GAIN_43dB	= 0x60,
			GAIN_48dB	= 0x70
		};
		
		
		//Struktura pro ulozeni UID karty
		typedef struct
		{
			uint8_t	size;	//Velikost UID (4, 7, 10 bajtu)
			uint8_t uid_data[10];
			uint8_t sak;	//Navratovy kod karty po uspesnem odeslani UID
		} UID;
		
		
		//Struktura pro MIFAFE Crypto1 klic
		typedef struct
		{
			uint8_t key_data[MF_KEY_SIZE];
		} mifare_key;

	private:
		spi_class &spi;
			
	//Promene
	public:
		
		//Inicializace UID
		UID uid;

	private:
		
		//Velikost MFRC522 FIFO
		static const uint8_t fifo_size = 64;
		
		//Port a pin resetovaciho pinu MFRC522
		uint8_t reset_port;
		uint8_t reset_pin;
		
		//Port a pin SS pinu MFRC522
		uint8_t select_port;
		uint8_t select_pin;
		
	
	//Metody
	public:
		
		//Konstruktor, destruktor
		MFRC522_class(spi_class &spi);
		MFRC522_class(spi_class &spi, uint8_t _select_port, uint8_t _select_pin, uint8_t _reset_port, uint8_t _reset_pin);
		~MFRC522_class();
		
		//Zakladni interface pro komunikaci s MFRC522
		void	reader_write_register(uint8_t _register, uint8_t _value);
		void	reader_write_register(uint8_t _register,uint8_t _count, uint8_t *_values);
		
		uint8_t reader_read_register(uint8_t _register);
		void	reader_read_register(uint8_t _register, uint8_t _count, uint8_t *_values, uint8_t _rx_align = 0);
		
		void	set_bit_mask(uint8_t _register, uint8_t _mask);
		
		void	reader_set_register_bit_mask(uint8_t _register, uint8_t _mask);
		void	reader_clear_register_bit_mask(uint8_t _register, uint8_t _mask);
		
		uint8_t reader_calculate_crc(uint8_t *_data, uint8_t _lenght, uint8_t *_result);
		
		
		//Metody pro obsluhu MFRC522
		void	reader_ini(void);
		void	reader_reset(void);
		void	reader_antena_on(void);
		void	reader_set_antena_gain(uint8_t _gain);
		
		
		//Metody pro komunikaci s rfid kartami
		uint8_t	reader_transmit_and_read_data(uint8_t *_send_data, uint8_t _send_lenght, uint8_t *_back_data, uint8_t *_back_lenght, uint8_t *_valid_bits = NULL, uint8_t _rx_align = 0, bool _check_crc = false);
		uint8_t reader_communicate_with_card(uint8_t _command, uint8_t _wait_irq, uint8_t *_send_data, uint8_t _send_lenght, uint8_t *_back_data = NULL, uint8_t *_back_lenght = NULL, uint8_t *_valid_bits = NULL, uint8_t _rx_align = 0, bool _check_crc = false);
		
		uint8_t card_request(uint8_t *_buffer_ATQA, uint8_t *_buffer_size);
		uint8_t card_wake_up(uint8_t *_buffer_ATQA, uint8_t *_buffer_size);
		uint8_t card_REQA_or_WUPA(uint8_t _command, uint8_t *_buffer_ATQA, uint8_t *_buffer_size);
		uint8_t card_select(UID *uid, uint8_t _valid_bits = 0);
		uint8_t card_halt(void);
		
		
		//Metody pro komunikaci s MIFARE kartami
		uint8_t	reader_authenticate(uint8_t _command, uint8_t _block_addr, mifare_key *key, UID *uid);
		void	reader_stop_crypto(void);
		uint8_t mifare_read(uint8_t _block_addr, uint8_t *_buffer, uint8_t *_buffer_size);
		uint8_t mifare_write(uint8_t _block_addr, uint8_t *_buffer, uint8_t _buffer_size);
		uint8_t mifare_decrement(uint8_t _block_addr, long _delta);
		uint8_t mifare_increment(uint8_t _block_addr, long _delta);
		uint8_t mifare_restore(uint8_t _block_addr);
		uint8_t mifare_transfer(uint8_t _block_addr);
		uint8_t mifare_ultralight_write(uint8_t _page, uint8_t *_buffer, uint8_t _buffer_size);
		
	
		//Podpurne metody
		uint8_t	reader_mifare_transmit_and_receive(uint8_t *_send_data, uint8_t _send_lenght, bool _accept_timeout = false);
		const char *get_status_code_name(uint8_t _code);
		uint8_t card_get_type(uint8_t _sak);
		const char *card_get_type_name(uint8_t _type);
		void	mifare_set_acces_bits(uint8_t *_acces_bit_buffer, uint8_t _g0, uint8_t _g1, uint8_t _g2, uint8_t _g3);
		
		
		//Uzitecne metody
		bool	card_is_new_present(void);
		bool	card_read_serial(void);
		
		//Odeslani dat na uart
		
		//Odeslani dat na lcd
		
		
	protected:
	private:
		uint8_t mifare_two_step_helper(uint8_t _command, uint8_t _block_addr, long _data);

}; //MFRC522_class



//(c) Martin Vitek 2013 (OK1KVK)
