#include "sht11.h"

#include "wait.h"

SHT11::SHT11(io *sda, io *sck)
{
	this->sda = sda;
	this->sck = sck;

	sck->set_direction(sck->OUTPUT);
	sda->set_direction(sda->OUTPUT);

	sda->on();
	sck->on();
}

void SHT11::start_tx()
{
	sda->set_direction(sda->OUTPUT);
	sda->on();
	wait_us(20);

	sck->off();
	wait_us(1);

	sck->on();
	wait_us(1);

	sda->off();
	wait_us(1);

	sck->off();
	wait_us(1);

	sck->on();
	wait_us(1);

	sda->on();
	wait_us(1);

	sck->off();
	wait_us(1);
}

void SHT11::send_command(commands command)
{
	for (uint8_t i=0; i<8; i++)
	{
		if (command & (1<<(7-i))) sda->on();
		else sda->off();
		wait_us(1);

		sck->on();
		wait_us(1);

		sck->off();
		if (i == 7)	sda->on();
		else sda->off();
		wait_us(1);
	}

	sda->set_direction(sda->INPUT);
	sda->off();
	wait_us(1);

	while (sda->is_set()) ;

	sck->on();
	wait_us(1);
	sck->off();
	wait_us(1);

	//while (!sda->is_set()) ;
}

uint16_t SHT11::read_data()
{
	while (sda->is_set()) ;

	uint16_t result(0);

	for (uint8_t i=0; i<8; i++)
	{
		sck->on();
		wait_us(1);
		if (sda->is_set()) result |= (1<<(15-i));
		sck->off();
		wait_us(1);
	}

	sda->set_direction(sda->OUTPUT);
	sda->off();
	sck->on();
	wait_us(1);

	sck->off();
	sda->set_direction(sda->INPUT);
	wait_us(1);

	for (uint8_t i=0; i<8; i++)
	{
		sck->on();
		wait_us(1);
		if (sda->is_set()) result |= (1<<(7-i));
		sck->off();
		wait_us(1);
	}

	return result;
}

void SHT11::stop_tx()
{
	sda->on();
	sda->set_direction(sda->OUTPUT);
	sda->on();
	wait_us(1);

	sck->on();
	wait_us(1);
	sck->off();
	wait_us(1);
}

uint16_t SHT11::measure_temperature()
{
	start_tx();
	send_command(TEMPERATURE);

	uint16_t result = read_data();

	stop_tx();

	return result;
}

uint16_t SHT11::measure_humidity()
{
	start_tx();
	send_command(HUMIDITY);

	uint16_t result = read_data();

	stop_tx();

	return result;
}

void SHT11::initialize()
{

}

void SHT11::communication_reset()
{
	sda->set_direction(sda->OUTPUT);
	sda->on();

	for (uint8_t i=0; i<10; i++)
	{
		sck->on();
		wait_us(1);
		sck->off();
		wait_us(1);
	}
}

[[deprecated]]
void SHT11::reset()
{
	send_command(SOFT_RESET);
}

bool SHT11::check_communication()
{
	send_command(READ_STATUS_REG);

	while (sda->is_set()) ;

	uint8_t result(0);

	for (uint8_t i=0; i<8; i++)
	{
		sck->on();
		wait_us(1);
		if (sda->is_set()) result |= (1<<(7-i));
		sck->off();
		wait_us(1);
	}

	stop_tx();

	return (bool)result;
}

void SHT11::start_measuring_temperature()
{
	start_tx();
	send_command(TEMPERATURE);
	measuring_temp = true;
}

void SHT11::start_measuring_humidity()
{
	start_tx();
	send_command(HUMIDITY);
	measuring_humi = true;
}

void SHT11::handler()
{
	if (measuring_temp && sda->is_set())
	{
		measuring_temp = false;

		temp_raw = read_data();
		stop_tx();

		temp_done = true;
	}

	if (measuring_humi && sda->is_set())
	{
		measuring_humi = false;
		humi_raw = read_data();
		stop_tx();

		humi_done = true;
	}
}

float SHT11::get_temperature()
{
	temperature = (-40.1+0.01*temp_raw);
	return temperature;
}

float SHT11::get_humidity()
{
	float humi_linear = (-2.0468+0.0367*humi_raw+(-1.5955e-6)*humi_raw*humi_raw);
	humidity = (temperature - 25.)*(0.01+0.00008*humi_raw)+humi_linear;

	return humidity;
}
